import os
import platform
import site
import sys
import subprocess


def find_env_setting(arg, env_var_name):
    # We check sys.argv for the arg, and os.environ for the env_var_name
    # The argument gets precedence
    if arg in sys.argv:
        return os.path.abspath(sys.argv[sys.argv.index(arg) + 1])

    if os.environ.get(env_var_name):
        return os.path.abspath(os.environ.get(env_var_name))


def prepend_python_path(root_dir, search_paths_dict):
    # root_dir is the root directory which we will search
    # search_paths_dict is a dictionary with "Linux", "Darwin" (Mac),
    # and "Windows" keys, with niladic functions for the values that
    # returns a list of the paths to search. The paths for the current
    # OS will be used.

    # Returns True on success, False on failure

    # Make sure the python lib dir can be found
    python_lib_path = find_python_path(root_dir, search_paths_dict)
    if python_lib_path is None:
        # A warning should be printed already from find_python_path()
        return False

    # Use site.addsitedir() to automatically process .pth files.
    # This also appends the paths to sys.path. We will move the new
    # paths to the front of sys.path.
    prev_length = len(sys.path)
    site.addsitedir(python_lib_path)
    sys.path[:] = sys.path[prev_length:] + sys.path[0:prev_length]

    return True


def append_path_to_environ(var_name, paths):
    result_paths = []
    current_var_value = os.environ.get(var_name, "")
    sys_path_sepa = ";" if platform.system() == "Windows" else ":"

    if len(current_var_value):
        result_paths.append(current_var_value)

    for path_entry in paths:
        if path_entry not in current_var_value:
            result_paths.append(path_entry)

    if len(result_paths):
        os.environ[var_name] = sys_path_sepa.join(result_paths)


def filter_existing_paths(root_dir, paths_to_try):
    found_paths = []

    for path in paths_to_try:
        full_path = os.path.join(root_dir, path)
        if os.path.exists(full_path):
            found_paths.append(full_path)

    return found_paths


def find_python_path(root_dir, search_paths_dict):
    # Returns the python path, or None if it was not found

    if not os.path.exists(root_dir):
        print(f"Warning: path '{root_dir}' does not exist")
        return None

    if platform.system() not in search_paths_dict:
        raise Exception(f"Unhandled system: {platform.system()}")

    paths_to_try = search_paths_dict[platform.system()]()
    found_paths = filter_existing_paths(root_dir, paths_to_try)

    if len(found_paths) != 1:
        paths_str = "\n".join(paths_to_try)
        print(
            "Warning: python library path could not be found in "
            f"'{root_dir}'. Tried:\n{paths_str}\n"
        )
        return None

    return found_paths[0]


def rerun(base_path, add_path_vars, remove_vars=[]):
    for name, paths in add_path_vars.items():
        resolved_paths = filter_existing_paths(base_path, paths)
        append_path_to_environ(name, resolved_paths)

    clear_environ_variables(*remove_vars)
    rerun_with_new_environ(*list(add_path_vars.keys()))


def rerun_with_new_environ(*var_to_print):
    if not os.environ.get("__IN_TRAME_RERUN") == "YES":
        env = os.environ.copy()
        env["__IN_TRAME_RERUN"] = "YES"

        # Print env info
        if len(var_to_print):
            print("-" * 80)
            print("Re-excuting with following environment variables:")
            for name in var_to_print:
                print(f" - {name}={env.get(name)}")
            print("-" * 80)

        # Re-run the same command with the modified environment
        cmd = [sys.executable] + sys.argv
        subprocess.run(cmd, stdout=sys.stdout, stderr=sys.stderr, env=env)
        sys.exit()


def clear_environ_variables(*names):
    for var in names:
        if var in os.environ:
            os.environ.pop(var)



from trame_server import Server
from trame_client import module
from trame_client.widgets.core import VirtualNode

# Ensure this is imported so that mimetypes.init() is decorated
import trame.app.mimetypes  # noqa: F401

DEFAULT_NAME = "trame"
AVAILABLE_SERVERS = {}


def get_server(name=None, create_if_missing=True, **kwargs):
    """
    Return a server for serving trame applications.
    If a name is given and such server is not available yet,
    it will be created otherwise the previously created instance will be returned.

    :param name: A server name identifier which can be useful when several servers
                are expected to be created. Most of the time, passing no arguments
                is what you are looking for.
    :type name: str


    :param create_if_missing: By default if a server for a given name does not exist
        that method will create it.
    :type create_if_missing: bool

    :param **kwargs: any extra keyword args are passed as option to the server instance.

    :return: Return a unique Server instance per given name.
    :rtype: trame_server.core.Server
    """
    if name is None:
        name = DEFAULT_NAME

    if name in AVAILABLE_SERVERS:
        return AVAILABLE_SERVERS[name]

    if create_if_missing:
        server = Server(name, VirtualNode, **kwargs)
        server.enable_module(module)  # Always load html module first
        AVAILABLE_SERVERS[name] = server
        return server

    # No server available for given name
    return None


__all__ = [
    "get_server",
]



import os
from urllib.error import HTTPError
from urllib.request import urlretrieve


def download_file_from_google_drive(id, destination):
    import requests

    URL = "https://docs.google.com/uc"

    session = requests.Session()
    response = session.get(
        URL, params={"id": id, "confirm": "t", "export": "download"}, stream=True
    )
    token = get_confirm_token(response)

    if token:
        params = {"id": id, "confirm": token}
        response = session.get(URL, params=params, stream=True)

    save_response_content(response, destination)


def get_confirm_token(response):
    for key, value in response.cookies.items():
        if key.startswith("download_warning"):
            return value

    return None


def save_response_content(response, destination):
    CHUNK_SIZE = 32768

    with open(destination, "wb") as f:
        for chunk in response.iter_content(CHUNK_SIZE):
            if chunk:  # filter out keep-alive new chunks
                f.write(chunk)


class AbstractRemoteFile:
    """
    AbstractRemoteFile provide infrastructure for RemoteFile where
    only the method fetch() needs to be defined for a concreate implementation.
    """

    def __init__(self, local_path=None, local_base=None):
        # setup base path
        self._base = os.getcwd()
        if local_base is not None:
            if os.path.exists(local_base):
                if os.path.isfile(local_base):
                    self._base = os.path.abspath(os.path.dirname(local_base))
                else:
                    self._base = os.path.abspath(local_base)
            else:
                self._base = os.path.abspath(local_base)

        # setup local path
        self._file_path = local_path
        if not os.path.isabs(local_path):
            self._file_path = os.path.abspath(os.path.join(self._base, local_path))

        # Make sure target directory exists
        parent_dir = os.path.dirname(self._file_path)
        if not os.path.exists(parent_dir):
            os.makedirs(parent_dir)

    @property
    def local(self):
        """Return true if the file is available locally on the File System"""
        return os.path.exists(self._file_path)

    def fetch(self):
        """Perform the action needed to fetch the content and store it locally"""
        pass

    @property
    def path(self):
        """Return the actual local file path"""
        if not self.local:
            self.fetch()

        return self._file_path


class GoogleDriveFile(AbstractRemoteFile):
    """
    Helper file to manage caching and retrieving of file available on Google Drive
    """

    def __init__(self, local_path=None, google_id=None, local_base=None):
        """
        Provide the information regarding where the file should be located
        and where to fetch it if missing.

        :param local_path: relative or absolute path
        :param google_id: Resource ID from google
        :param local_base: Absolute path when local_path is relative
        """
        super().__init__(local_path, local_base)
        self._gid = google_id

    def fetch(self):
        try:
            print(f"Downloading:\n - {self._gid}\n - to {self._file_path}")
            download_file_from_google_drive(self._gid, self._file_path)
        except HTTPError as e:
            print(RuntimeError(f"Failed to download {self._gid}. {e.reason}"))


class HttpFile(AbstractRemoteFile):
    """
    Helper file to manage caching and retrieving of file available on HTTP servers
    """

    def __init__(self, local_path=None, remote_url=None, local_base=None):
        """
        Provide the information regarding where the file should be located
        and where to fetch it if missing.

        :param local_path: relative or absolute path
        :param remote_url: http(s):// url to fetch the file from
        :param local_base: Absolute path when local_path is relative
        """
        super().__init__(local_path, local_base)
        self._url = remote_url

    def fetch(self):
        try:
            print(f"Downloading:\n - {self._url}\n - to {self._file_path}")
            urlretrieve(self._url, self._file_path)
        except HTTPError as e:
            print(RuntimeError(f"Failed to download {self._url}. {e.reason}"))


__all__ = [
    "AbstractRemoteFile",
    "GoogleDriveFile",
    "HttpFile",
]



# -*- coding: utf-8 -*-
"""Set up sys.path to include ParaView, if available

This checks the TRAME_PARAVIEW environment variable and the --paraview
argument in sys.argv. The specified path will then be automatically
searched for the python library directory and, if found, it will be
added to the sys.path so that the ParaView libraries may be imported.

To use: specify one of the following, and then import this file:
 - `--paraview /path/to/paraview/release` argument
 - environment variable `TRAME_PARAVIEW=/path/to/paraview/release`
"""
import os
import platform
import sys

from . import utils

PV_LOADED = False

PV_HOME = utils.find_env_setting("--paraview", "TRAME_PARAVIEW")

RAISE_ERRORS = os.getenv("TRAME_PARAVIEW_FAIL_SILENTLY") is None

if PV_HOME is None and RAISE_ERRORS:
    msg = (
        "trame.env.paraview was imported, but the paraview location was not "
        "defined. Define it with either the argument '--paraview <path>' or "
        "the TRAME_PARAVIEW environment variable"
    )
    raise Exception(msg)


# Each of the following functions returns a list of paths to try
# for the specified OS.
def linux_paths():
    py_major, py_minor = sys.version_info[:2]
    return [f"lib/python{py_major}.{py_minor}/site-packages"]


def windows_paths():
    return ["bin/Lib/site-packages"]


search_paths_dict = {
    "Linux": linux_paths,
    "Windows": windows_paths,
}

if PV_HOME:
    if platform.system() == "Darwin":
        env_paths = {
            "PYTHONPATH": ["Contents/Python"],
            "DYLD_LIBRARY_PATH": ["Contents/Libraries"],
        }
        utils.rerun(PV_HOME, env_paths, ["PV_VENV", "VTK_VENV"])
    else:
        PV_LOADED = utils.prepend_python_path(PV_HOME, search_paths_dict)
        if not PV_LOADED and RAISE_ERRORS:
            raise Exception(f"Failed to add paraview python libraries at {PV_HOME}")



r"""
From the directory containing the static content for a trame application to work,
generate another application specific HTML file.
"""
import argparse
import re
from pathlib import Path
import sys

APP_PATTERN = re.compile(r'data-app-name="\w+"')


def create_app_file(input_file, output_file, app_name):
    # Read in the file
    with open(input_file, "r") as f_in:
        content = f_in.read()
        patched_content = APP_PATTERN.sub(f'data-app-name="{app_name}"', content)
        with open(output_file, "w") as f_out:
            f_out.write(patched_content)


def main():
    parser = argparse.ArgumentParser(
        description="HTML app file generator for trame applications"
    )

    parser.add_argument(
        "--name",
        default="trame",
        help="Application name to encode inside HTML {name}.html",
    )

    parser.add_argument(
        "--input",
        help="Input file to use as template",
    )

    args, _ = parser.parse_known_args()

    # Handle input
    input_file = Path(args.input)
    if not input_file.exists():
        parser.print_help()
        sys.exit(0)

    if input_file.is_dir():
        input_file = input_file / "index.html"

    if not input_file.exists():
        parser.print_help()
        sys.exit(0)

    output_file = input_file.with_stem(args.name)
    create_app_file(input_file, output_file, args.name)


if __name__ == "__main__":
    main()



from typing import Type, TypeVar, Generic

T = TypeVar("T")


class Singleton(Generic[T]):
    """
    Class decorator to make it a Singleton

    This is useful when you want a central engine instance or else
    to be used across your application modules.

    But using such decorator will make it tricky or impossible to
    use your application within several server using the same
    event loop.
    """

    def __init__(self, cls: Type[T]):
        self._instance: T = cls()

    def __call__(self) -> T:
        return self._instance


__all__ = [
    "Singleton",
]
